/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef CSR_SPMV_PART_VECTOR_GPU_H
#define CSR_SPMV_PART_VECTOR_GPU_H

// -----------------------------------------------------------------------------

#include "csr_matrix_data.h"
#include "operator_T.h"
#include <cstddef>
#include <algorithm> // std::min
#include <cuda_runtime.h>
#include <thrust/detail/device/cuda/arch.h>

// -----------------------------------------------------------------------------

const unsigned int CSR_SPMV_PART_VECTOR_GPU_NUM_VORDER = 2;

// -----------------------------------------------------------------------------

static const unsigned int CSR_SPMV_PART_THREADS_PER_BLOCK = 128;

// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_csr_vector_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2010 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------

template
<
	typename index_type,
	typename value_type,
	unsigned int VECTORS_PER_BLOCK,
	unsigned int THREADS_PER_VECTOR,
	typename assign_add_subtract
>
__global__ void
csr_spmv_part_vector_vorder_0_kernel(const index_type* Ap, // A.row_offsets
                                     const index_type* Aj, // A.column_indices
                                     const value_type* Ax, // A.values
                                     const value_type* x, 
                                           value_type* y,
                                     index_type part_row_start,
                                     index_type part_row_end,
                                     index_type part_col_start,
                                     index_type part_col_end,
                                     index_type num_vecs,
                                     index_type next_vec_offset_x,
                                     index_type next_vec_offset_y)
{
	__shared__ volatile value_type sdata[VECTORS_PER_BLOCK * THREADS_PER_VECTOR + THREADS_PER_VECTOR / 2];  // padded to avoid reduction conditionals
	__shared__ volatile index_type ptrs[VECTORS_PER_BLOCK][2];

	const index_type thread_id   = CSR_SPMV_PART_THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;    // global thread index
	const index_type thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);          // thread index within the vector
	const index_type vector_id   = thread_id   /  THREADS_PER_VECTOR;               // global vector index
	const index_type vector_lane = threadIdx.x /  THREADS_PER_VECTOR;               // vector index within the block
	const index_type num_vectors = VECTORS_PER_BLOCK * gridDim.x;                   // total number of active vectors

	for(index_type ivec = 0; ivec < num_vecs; ++ivec, x += next_vec_offset_x, y += next_vec_offset_y)
	{
		for(index_type row = vector_id + part_row_start; row < part_row_end; row += num_vectors)
		{
			// use two threads to fetch Ap[row] and Ap[row+1]
			// this is considerably faster than the straightforward version
			if(thread_lane < 2)
				ptrs[vector_lane][thread_lane] = Ap[row + thread_lane];

			const index_type row_start = ptrs[vector_lane][0];                   //same as: row_start = Ap[row];
			const index_type row_end   = ptrs[vector_lane][1];                   //same as: row_end   = Ap[row+1];

			// initialize local sum
			value_type sum = 0;

			if(THREADS_PER_VECTOR == 32 && row_end - row_start > 32)
			{
				// ensure aligned memory access to Aj and Ax

				index_type jj = row_start - (row_start & (THREADS_PER_VECTOR - 1)) + thread_lane;

				// accumulate local sums
				if(jj >= row_start && jj < row_end && part_col_start <= Aj[jj] && Aj[jj] < part_col_end)
				{
					sum += Ax[jj] * x[Aj[jj] - part_col_start];
				}

				// accumulate local sums
				for(jj += THREADS_PER_VECTOR; jj < row_end; jj += THREADS_PER_VECTOR)
				{
					if(part_col_start <= Aj[jj] && Aj[jj] < part_col_end)
						sum += Ax[jj] * x[Aj[jj] - part_col_start];
				}
			}
			else
			{
				// accumulate local sums
				for(index_type jj = row_start + thread_lane; jj < row_end; jj += THREADS_PER_VECTOR)
					if(part_col_start <= Aj[jj] && Aj[jj] < part_col_end)
						sum += Ax[jj] * x[Aj[jj] - part_col_start];
			}

			// store local sum in shared memory
			sdata[threadIdx.x] = sum;

			// reduce local sums to row sum
			if(THREADS_PER_VECTOR > 16) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x + 16];
			if(THREADS_PER_VECTOR >  8) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  8];
			if(THREADS_PER_VECTOR >  4) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  4];
			if(THREADS_PER_VECTOR >  2) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  2];
			if(THREADS_PER_VECTOR >  1) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  1];

			// first thread writes the result
			if(thread_lane == 0)
				assign_add_subtract::op(y[row - part_row_start], sdata[threadIdx.x]);
		}
	}
}

// -----------------------------------------------------------------------------

template
<
	typename index_type,
	typename value_type,
	unsigned int VECTORS_PER_BLOCK,
	unsigned int THREADS_PER_VECTOR,
	typename assign_add_subtract
>
__global__ void
csr_spmv_part_vector_vorder_1_kernel(const index_type* Ap, // A.row_offsets
                                     const index_type* Aj, // A.column_indices
                                     const value_type* Ax, // A.values
                                     const value_type* x, 
                                           value_type* y,
                                     index_type part_row_start,
                                     index_type part_row_end,
                                     index_type part_col_start,
                                     index_type part_col_end,
                                     index_type num_vecs,
                                     index_type next_vec_offset_x,
                                     index_type next_vec_offset_y)
{
	__shared__ volatile value_type sdata[VECTORS_PER_BLOCK * THREADS_PER_VECTOR + THREADS_PER_VECTOR / 2];  // padded to avoid reduction conditionals
	__shared__ volatile index_type ptrs[VECTORS_PER_BLOCK][2];

	const index_type thread_id   = CSR_SPMV_PART_THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;    // global thread index
	const index_type thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);          // thread index within the vector
	const index_type vector_id   = thread_id   /  THREADS_PER_VECTOR;               // global vector index
	const index_type vector_lane = threadIdx.x /  THREADS_PER_VECTOR;               // vector index within the block
	const index_type num_vectors = VECTORS_PER_BLOCK * gridDim.x;                   // total number of active vectors

	for(index_type row = vector_id + part_row_start; row < part_row_end; row += num_vectors,
		x -= num_vecs * next_vec_offset_x, y -= num_vecs * next_vec_offset_y)
	{
		// use two threads to fetch Ap[row] and Ap[row+1]
		// this is considerably faster than the straightforward version
		if(thread_lane < 2)
			ptrs[vector_lane][thread_lane] = Ap[row + thread_lane];

		const index_type row_start = ptrs[vector_lane][0];                   //same as: row_start = Ap[row];
		const index_type row_end   = ptrs[vector_lane][1];                   //same as: row_end   = Ap[row+1];

		for(index_type ivec = 0; ivec < num_vecs; ++ivec, x += next_vec_offset_x, y += next_vec_offset_y)
		{
			// initialize local sum
			value_type sum = 0;

			if(THREADS_PER_VECTOR == 32 && row_end - row_start > 32)
			{
				// ensure aligned memory access to Aj and Ax

				index_type jj = row_start - (row_start & (THREADS_PER_VECTOR - 1)) + thread_lane;

				// accumulate local sums
				if(jj >= row_start && jj < row_end && part_col_start <= Aj[jj] && Aj[jj] < part_col_end)
				{
					sum += Ax[jj] * x[Aj[jj] - part_col_start];
				}

				// accumulate local sums
				for(jj += THREADS_PER_VECTOR; jj < row_end; jj += THREADS_PER_VECTOR)
				{
					if(part_col_start <= Aj[jj] && Aj[jj] < part_col_end)
						sum += Ax[jj] * x[Aj[jj] - part_col_start];
				}
			}
			else
			{
				// accumulate local sums
				for(index_type jj = row_start + thread_lane; jj < row_end; jj += THREADS_PER_VECTOR)
					if(part_col_start <= Aj[jj] && Aj[jj] < part_col_end)
						sum += Ax[jj] * x[Aj[jj] - part_col_start];
			}

			// store local sum in shared memory
			sdata[threadIdx.x] = sum;

			// reduce local sums to row sum
			if(THREADS_PER_VECTOR > 16) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x + 16];
			if(THREADS_PER_VECTOR >  8) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  8];
			if(THREADS_PER_VECTOR >  4) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  4];
			if(THREADS_PER_VECTOR >  2) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  2];
			if(THREADS_PER_VECTOR >  1) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  1];

			// first thread writes the result
			if(thread_lane == 0)
				assign_add_subtract::op(y[row - part_row_start], sdata[threadIdx.x]);
		}
	}
}

// -----------------------------------------------------------------------------

template<typename index_type>
static inline index_type divide_into(index_type x, index_type y)
{
	// ceil(x/y) for integers, used to determine # of blocks/warps etc.
	return (x + y - 1)/y;
}

// -----------------------------------------------------------------------------

template
<
	unsigned int THREADS_PER_VECTOR,
	typename assign_add_subtract,
	typename index_type,
	typename value_type
>
static void csr_spmv_part_vector_gpu_T(const csr_matrix_data<index_type, value_type>& A, 
                                       index_type part_row_start,
                                       index_type part_row_end,
                                       index_type part_col_start,
                                       index_type part_col_end,
                                       const value_type* x, 
                                       value_type* y,
                                       index_type num_vecs,
                                       index_type next_vec_offset_x,
                                       index_type next_vec_offset_y,
                                       unsigned int vorder)
{
	const unsigned int VECTORS_PER_BLOCK  = CSR_SPMV_PART_THREADS_PER_BLOCK / THREADS_PER_VECTOR;

	if(vorder == 0)
	{
		const unsigned int MAX_BLOCKS = static_cast<unsigned int>(thrust::detail::device::cuda::arch::max_active_blocks(csr_spmv_part_vector_vorder_0_kernel<
			index_type, value_type, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, assign_add_subtract>,
			CSR_SPMV_PART_THREADS_PER_BLOCK, (size_t) 0));

		const unsigned int NUM_BLOCKS = std::min(MAX_BLOCKS,
			static_cast<unsigned int>(divide_into(part_row_end - part_row_start, index_type(VECTORS_PER_BLOCK))));

		csr_spmv_part_vector_vorder_0_kernel<index_type, value_type, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, assign_add_subtract>
			<<<NUM_BLOCKS, CSR_SPMV_PART_THREADS_PER_BLOCK>>> 
			(A.row_offsets, A.column_indices, A.values, x, y, part_row_start, part_row_end,
			part_col_start, part_col_end, num_vecs, next_vec_offset_x, next_vec_offset_y);
	}
	else if(vorder == 1)
	{
		const unsigned int MAX_BLOCKS = static_cast<unsigned int>(thrust::detail::device::cuda::arch::max_active_blocks(csr_spmv_part_vector_vorder_1_kernel<
			index_type, value_type, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, assign_add_subtract>,
			CSR_SPMV_PART_THREADS_PER_BLOCK, (size_t) 0));

		const unsigned int NUM_BLOCKS = std::min(MAX_BLOCKS,
			static_cast<unsigned int>(divide_into(part_row_end - part_row_start, index_type(VECTORS_PER_BLOCK))));

		csr_spmv_part_vector_vorder_1_kernel<index_type, value_type, VECTORS_PER_BLOCK, THREADS_PER_VECTOR, assign_add_subtract>
			<<<NUM_BLOCKS, CSR_SPMV_PART_THREADS_PER_BLOCK>>> 
			(A.row_offsets, A.column_indices, A.values, x, y, part_row_start, part_row_end,
			part_col_start, part_col_end, num_vecs, next_vec_offset_x, next_vec_offset_y);
	}
}

// -----------------------------------------------------------------------------

template<typename assign_add_subtract, typename index_type, typename value_type>
bool csr_spmv_part_vector_gpu(const csr_matrix_data<index_type, value_type>& A, 
                                     index_type part_row_start,
                                     index_type part_row_end,
                                     index_type part_col_start,
                                     index_type part_col_end,
                                     index_type part_num_entries,
                                     const value_type* x, 
                                     value_type* y,
                                     index_type num_vecs,
                                     index_type next_vec_offset_x,
                                     index_type next_vec_offset_y,
                                     unsigned int vorder)
{
	bool success = false;

	if( part_row_start <= part_row_end &&
		part_row_end <= A.num_rows &&
		part_col_start <= part_col_end &&
		part_col_end <= A.num_cols &&
		x &&
		y &&
		num_vecs > 0 &&
		next_vec_offset_x >= part_col_end - part_col_start &&
		next_vec_offset_y >= part_row_end - part_row_start &&
		vorder < CSR_SPMV_PART_VECTOR_GPU_NUM_VORDER)
	{
		const index_type nnz_per_row = part_num_entries / (part_row_end - part_row_start);

		if(     nnz_per_row <=  2) csr_spmv_part_vector_gpu_T< 2, assign_add_subtract>(A, part_row_start, part_row_end, part_col_start, part_col_end, x, y, num_vecs, next_vec_offset_x, next_vec_offset_y, vorder);
		else if(nnz_per_row <=  4) csr_spmv_part_vector_gpu_T< 4, assign_add_subtract>(A, part_row_start, part_row_end, part_col_start, part_col_end, x, y, num_vecs, next_vec_offset_x, next_vec_offset_y, vorder);
		else if(nnz_per_row <=  8) csr_spmv_part_vector_gpu_T< 8, assign_add_subtract>(A, part_row_start, part_row_end, part_col_start, part_col_end, x, y, num_vecs, next_vec_offset_x, next_vec_offset_y, vorder);
		else if(nnz_per_row <= 16) csr_spmv_part_vector_gpu_T<16, assign_add_subtract>(A, part_row_start, part_row_end, part_col_start, part_col_end, x, y, num_vecs, next_vec_offset_x, next_vec_offset_y, vorder);
		else                       csr_spmv_part_vector_gpu_T<32, assign_add_subtract>(A, part_row_start, part_row_end, part_col_start, part_col_end, x, y, num_vecs, next_vec_offset_x, next_vec_offset_y, vorder);

		success = true;
	 }

	 return success;
}

// -----------------------------------------------------------------------------

// return value true means successful, else not.  A, x, y reside in GPU memory.
// A points to the "full" matrix.  "x" points to the only part of the memory
// that will be used in multiplying.  Same for "y".  What this means is that
// x need not point to memory of size A.num_cols and y need not point to memory
// size A.num_rows.  The minimal is needed.  For x that means part_col_end - part_col_start
// and for y, part_row_end - part_row_start.

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
bool csr_spmv_assign_part_vector_gpu(const csr_matrix_data<index_type, value_type>& A, 
                                     index_type part_row_start,
                                     index_type part_row_end,
                                     index_type part_col_start,
                                     index_type part_col_end,
                                     index_type part_num_entries,
                                     const value_type* x,
                                     value_type* y,
                                     index_type num_vecs,
                                     index_type next_vec_offset_x,
                                     index_type next_vec_offset_y,
                                     unsigned int vorder)
{
	return csr_spmv_part_vector_gpu<operator_assign>(A, part_row_start, part_row_end, part_col_start,
		part_col_end, part_num_entries, x, y, num_vecs, next_vec_offset_x, next_vec_offset_y, vorder);
}

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
bool csr_spmv_add_part_vector_gpu(const csr_matrix_data<index_type, value_type>& A, 
                                  index_type part_row_start,
                                  index_type part_row_end,
                                  index_type part_col_start,
                                  index_type part_col_end,
                                  index_type part_num_entries,
                                  const value_type* x,
                                  value_type* y,
                                  index_type num_vecs,
                                  index_type next_vec_offset_x,
                                  index_type next_vec_offset_y,
                                  unsigned int vorder)
{
	return csr_spmv_part_vector_gpu<operator_add>(A, part_row_start, part_row_end, part_col_start,
		part_col_end, part_num_entries, x, y, num_vecs, next_vec_offset_x, next_vec_offset_y, vorder);
}

// -----------------------------------------------------------------------------

template<typename index_type, typename value_type>
bool csr_spmv_sub_part_vector_gpu(const csr_matrix_data<index_type, value_type>& A, 
                                  index_type part_row_start,
                                  index_type part_row_end,
                                  index_type part_col_start,
                                  index_type part_col_end,
                                  index_type part_num_entries,
                                  const value_type* x,
                                  value_type* y,
                                  index_type num_vecs,
                                  index_type next_vec_offset_x,
                                  index_type next_vec_offset_y,
                                  unsigned int vorder)
{
	return csr_spmv_part_vector_gpu<operator_sub>(A, part_row_start, part_row_end, part_col_start,
		part_col_end, part_num_entries, x, y, num_vecs, next_vec_offset_x, next_vec_offset_y, vorder);
}

// -----------------------------------------------------------------------------

#endif // CSR_SPMV_PART_VECTOR_GPU_H
