/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */

#ifndef CUDA_MEMORY_H
#define CUDA_MEMORY_H

// -----------------------------------------------------------------------------

#include <cstddef>
#include <algorithm>
#include <vector>
#include <stdexcept>
#include <cassert>
#include <cuda_runtime.h>
#include "cuda_error_utils.h"

// -----------------------------------------------------------------------------

template<typename type>
class cuda_memory
{
	// -------------------------------------------------------------------------

public:

	typedef type		value_type;
	typedef std::size_t	size_type;

	// -------------------------------------------------------------------------

	cuda_memory();

	cuda_memory(size_type N_size, const value_type* host_mem = 0);

	cuda_memory(const std::vector<value_type>& host_mem);

	cuda_memory(value_type* gpu_mem, size_type N_size, bool own_it);

	cuda_memory(const cuda_memory& other);

	// -------------------------------------------------------------------------

	virtual ~cuda_memory();

	// -------------------------------------------------------------------------

	// Use use_memory(...), disown(), and own() with care.

	bool use_memory(value_type* gpu_mem, size_type N_size, bool own_it);

	void disown();

	bool own();

	// -------------------------------------------------------------------------

	cuda_memory& operator=(const cuda_memory& other);

	bool deep_copy(const cuda_memory& other);

	bool resize(size_type N_size, const value_type* host_mem = 0);

	bool clear();

	// -------------------------------------------------------------------------

	bool zero();

	bool copy_from_host(const value_type* host_mem);

	bool copy_from_host(const std::vector<value_type>& vec);

	bool copy_to_host(value_type* host_mem);

	bool copy_to_host(std::vector<value_type>& vec);

	// -------------------------------------------------------------------------

	// offsets are not in bytes but in terms of number of value_type
	// offsets refer to offsets in gpu memory starting from m_store.

	// operations with std::vector<> assume that vector.size is larger or equal to
	// that is needed.  they'll work from the beginning of vector.

	bool copy_from_host(const value_type* host_mem, size_type offset_start, size_type offset_end);

	bool copy_from_host(const std::vector<value_type>& vec, size_type offset_start, size_type offset_end);

	bool copy_to_host(value_type* host_mem, size_type offset_start, size_type offset_end);

	bool copy_to_host(std::vector<value_type>& vec, size_type offset_start, size_type offset_end);

	// -------------------------------------------------------------------------

	// will not allocate, and will copy only if sizes match.
	bool copy_only_data(const cuda_memory& other);

	// -------------------------------------------------------------------------

	size_type size() const;

	size_type mem_size() const;

	bool same_size_as(const cuda_memory& other) const;

	value_type* gpu_mem_ptr();

	const value_type* gpu_mem_ptr() const;

	// -------------------------------------------------------------------------

	value_type at(size_type i) const;

	// -------------------------------------------------------------------------

protected:

	value_type*	m_store;
	size_type	m_N_size;
	bool		m_owned;

	// -------------------------------------------------------------------------
};


// -----------------------------------------------------------------------------

template<typename type>
cuda_memory<type>::cuda_memory()
	: m_store(0),
	  m_N_size(0),
	  m_owned(false)
{
}

template<typename type>
cuda_memory<type>::cuda_memory(size_type N_size, const value_type* host_mem)
	: m_store(0),
	  m_N_size(0),
	  m_owned(false)
{
	if(!resize(N_size, host_mem))
		throw std::runtime_error("cuda_memory: construction failed");
}

template<typename type>
cuda_memory<type>::cuda_memory(const std::vector<value_type>& host_mem)
	: m_store(0),
	  m_N_size(0),
	  m_owned(false)
{
	const typename std::vector<value_type>::size_type sz = host_mem.size();

	if(sz)
		if(!resize(sz, &host_mem.front()))
			throw std::runtime_error("cuda_memory: construction using std::vector<> failed");
}

template<typename type>
cuda_memory<type>::cuda_memory(value_type* gpu_mem, size_type N_size, bool own_it)
	: m_store(0),
	  m_N_size(0),
	  m_owned(false)
{
	if(!use_memory(gpu_mem, N_size, own_it))
		throw std::runtime_error("cuda_memory: cannot own given pointer");
}

template<typename type>
cuda_memory<type>::cuda_memory(const cuda_memory& other)
{
	deep_copy(other);
}

// -----------------------------------------------------------------------------

template<typename type>
cuda_memory<type>::~cuda_memory()
{
	bool success = clear();

	assert(success);

	// we don't throw here.
}

// -----------------------------------------------------------------------------

template<typename type>
bool cuda_memory<type>::use_memory(value_type* gpu_mem, size_type N_size, bool own_it)
{
	bool success = true;

	if(gpu_mem)
	{
		m_store  = gpu_mem;
		m_N_size = N_size;
		m_owned  = own_it;
	}
	else if(own_it)
	{
		success = false;
	}

	return success;
}

template<typename type>
void cuda_memory<type>::disown()
{
	m_owned = false;
}

template<typename type>
bool cuda_memory<type>::own()
{
	bool success = false;

	if(m_store)
	{
		m_owned = true;
		success = true;
	}

	return success;
}

// -----------------------------------------------------------------------------

template<typename type>
cuda_memory<type>& cuda_memory<type>::operator=(const cuda_memory& other)
{
	deep_copy(other);
	return *this;
}

template<typename type>
bool cuda_memory<type>::deep_copy(const cuda_memory& other)
{
	bool success = false;

	if(this != &other)
	{
		if(other.m_store)
		{
			if(other.m_owned)
			{
				cuda_memory tmp;
				success = tmp.resize(other.m_N_size);

				if(success)
				{
					cudaError_t cuda_ret =
						cudaMemcpy(tmp.m_store, other.m_store, other.mem_size(),
							cudaMemcpyDeviceToDevice);

					if(cuda_ret == cudaSuccess)
					{
						std::swap(m_store,  tmp.m_store);
						std::swap(m_N_size, tmp.m_N_size);
						std::swap(m_owned,  tmp.m_owned);
						success = true;
					}
					else
					{
						check_if_cuda_error(cuda_ret);
					}
				}
			}
			else
			{
				m_store  = other.m_store;
				m_N_size = other.m_N_size;
				m_owned  = false;
				success = true;
			}
		}
		else
		{
			success = clear();
		}
	}

	return success;
}

template<typename type>
bool cuda_memory<type>::resize(size_type N_size, const value_type* host_mem)
{
	bool success = false;

	// should not resize memory not owned (if any exists)
	if(! (m_store && !m_owned))
	{
		cuda_memory tmp;
		tmp.m_N_size = N_size;

		if(m_store && N_size == m_N_size)
		{
			// m_owned is true because of check above.
			tmp.m_store = m_store;
			tmp.m_owned = false;
		}
		else
		{
			// m_owned is false otherwise m_store cannot be 0.
			cuda_checked_call(cudaMalloc(&tmp.m_store, N_size * sizeof(value_type)));
			tmp.m_owned  = true;
		}

		success = (tmp.m_store != 0);

		if(success && host_mem)
		{
			cudaError_t cuda_ret =
				cudaMemcpy(tmp.m_store, host_mem,
					N_size * sizeof(value_type), cudaMemcpyHostToDevice);

			success = (cuda_ret == cudaSuccess);
		}

		if(success && tmp.m_owned)
		{
			std::swap(m_store,  tmp.m_store);
			std::swap(m_N_size, tmp.m_N_size);
			std::swap(m_owned,   tmp.m_owned);
		}
	}

	return success;
}

template<typename type>
bool cuda_memory<type>::clear()
{
	bool success = true;

	if(m_store && m_owned)
	{
		cudaError_t cuda_ret = cudaFree(m_store);

		if(cuda_ret == cudaSuccess)
		{
			m_store  = 0;
			m_N_size = 0;
			m_owned  = false;
		}
		else
		{
			success = false;
		}
	}

	return success;
}

// -----------------------------------------------------------------------------

template<typename type>
bool cuda_memory<type>::zero()
{
	bool success = false;

	if(m_store)
	{
		cudaError_t cuda_ret = cudaMemset(m_store, 0, mem_size());

		if(cuda_ret == cudaSuccess)
		{
			success = true;
		}
		else
		{
			check_if_cuda_error(cuda_ret);
		}
	}

	return success;
}

template<typename type>
bool cuda_memory<type>::copy_from_host(const value_type* host_mem)
{
	return copy_from_host(host_mem, 0, m_N_size);
}

template<typename type>
bool cuda_memory<type>::copy_from_host(const std::vector<value_type>& vec)
{
	return copy_from_host(vec, 0, m_N_size);
}

template<typename type>
bool cuda_memory<type>::copy_to_host(value_type* host_mem)
{
	return copy_to_host(host_mem, 0, m_N_size);
}

template<typename type>
bool cuda_memory<type>::copy_to_host(std::vector<value_type>& vec)
{
	return copy_to_host(vec, 0, m_N_size);
}

// -----------------------------------------------------------------------------

template<typename type>
bool cuda_memory<type>::copy_from_host(const value_type* host_mem, size_type offset_start, size_type offset_end)
{
	bool success = false;

	if(m_store && offset_start <= offset_end && offset_end <= m_N_size)
	{
		cudaError_t cuda_ret =
			cudaMemcpy(m_store + offset_start, host_mem + offset_start,
				sizeof(value_type) * (offset_end - offset_start), cudaMemcpyHostToDevice);

		if(cuda_ret == cudaSuccess)
		{
			success = true;
		}
		else
		{
			check_if_cuda_error(cuda_ret);
		}
	}

	return success;
}

template<typename type>
bool cuda_memory<type>::copy_from_host(const std::vector<value_type>& vec, size_type offset_start, size_type offset_end)
{
	bool success = false;

	if(m_store && vec.size() >= offset_end - offset_start &&
		offset_start <= offset_end && offset_end <= m_N_size)
	{
		cudaError_t cuda_ret =
			cudaMemcpy(m_store + offset_start, &vec.front() + offset_start,
				sizeof(value_type) * (offset_end - offset_start), cudaMemcpyHostToDevice);

		if(cuda_ret == cudaSuccess)
		{
			success = true;
		}
		else
		{
			check_if_cuda_error(cuda_ret);
		}
	}

	return success;
}

template<typename type>
bool cuda_memory<type>::copy_to_host(value_type* host_mem, size_type offset_start, size_type offset_end)
{
	bool success = false;

	if(m_store && offset_start <= offset_end && offset_end <= m_N_size)
	{
		cudaError_t cuda_ret =
			cudaMemcpy(host_mem, m_store + offset_start,
				sizeof(value_type) * (offset_end - offset_start), cudaMemcpyDeviceToHost);

		if(cuda_ret == cudaSuccess)
		{
			success = true;
		}
		else
		{
			check_if_cuda_error(cuda_ret);
		}
	}

	return success;
}

template<typename type>
bool cuda_memory<type>::copy_to_host(std::vector<value_type>& vec, size_type offset_start, size_type offset_end)
{
	bool success = false;

	if(m_store && vec.size() >= offset_end - offset_start &&
		offset_start <= offset_end && offset_end <= m_N_size)
	{
		cudaError_t cuda_ret =
			cudaMemcpy(&vec.front(), m_store + offset_start,
				sizeof(value_type) * (offset_end - offset_start), cudaMemcpyDeviceToHost);

		if(cuda_ret == cudaSuccess)
		{
			success = true;
		}
		else
		{
			check_if_cuda_error(cuda_ret);
		}
	}

	return success;
}

template<typename type>
bool cuda_memory<type>::copy_only_data(const cuda_memory& other)
{
	bool success = false;

	if(m_store && other.m_store && same_size_as(other))
	{
		cudaError_t cuda_ret =
			cudaMemcpy(m_store, other.m_store, mem_size(), cudaMemcpyDeviceToDevice);

		if(cuda_ret == cudaSuccess)
		{
			success = true;
		}
		else
		{
			check_if_cuda_error(cuda_ret);
		}
	}

	return success;
}

// -----------------------------------------------------------------------------

template<typename type>
typename cuda_memory<type>::size_type cuda_memory<type>::size() const
{
	return m_N_size;
}

template<typename type>
typename cuda_memory<type>::size_type cuda_memory<type>::mem_size() const
{
	return m_N_size * sizeof(value_type);
}

template<typename type>
bool cuda_memory<type>::same_size_as(const cuda_memory& other) const
{
	return other.m_N_size == m_N_size;
}

template<typename type>
typename cuda_memory<type>::value_type* cuda_memory<type>::gpu_mem_ptr()
{
	return m_store;
}

template<typename type>
const typename cuda_memory<type>::value_type* cuda_memory<type>::gpu_mem_ptr() const
{
	return m_store;
}

// -----------------------------------------------------------------------------

template<typename type>
typename cuda_memory<type>::value_type cuda_memory<type>::at(size_type i) const
{
	value_type ret;

	if(i < m_N_size && m_store)
	{
		cuda_checked_call(cudaMemcpy(&ret, m_store + i, sizeof(value_type),
			cudaMemcpyDeviceToHost));
	}
	else
	{
		throw std::out_of_range("cuda_memory::at: invalid subscript reading gpu memory");
	}

	return ret;
}

// -----------------------------------------------------------------------------

#endif // CUDA_MEMORY_H
